# ruff: noqa: N815
"""Defines the schema for the Assembly API."""

import functools
import logging
from dataclasses import dataclass
from enum import Enum
from typing import Any, Literal

import numpy as np
from pydantic import BaseModel

from onshape.onshape.schema.common import ElementUid

logger = logging.getLogger(__name__)

Key = tuple[str, ...]


@dataclass
class MimicRelation:
    parent: Key
    multiplier: float


class EmptyInstance(BaseModel):
    id: str
    name: str


class BaseInstance(BaseModel):
    name: str
    suppressed: bool
    id: str
    fullConfiguration: str
    configuration: str
    documentMicroversion: str
    documentId: str
    elementId: str

    @property
    def euid(self) -> ElementUid:
        return ElementUid(
            self.documentId,
            self.documentMicroversion,
            self.elementId,
            "",
            self.fullConfiguration,
        )


class AssemblyInstance(BaseInstance):
    type: Literal["Assembly"]


class PartInstance(BaseInstance):
    type: Literal["Part"]
    isStandardContent: bool
    partId: str

    @property
    def euid(self) -> ElementUid:
        return ElementUid(
            self.documentId,
            self.documentMicroversion,
            self.elementId,
            self.partId,
            self.fullConfiguration,
        )


Instance = AssemblyInstance | PartInstance | EmptyInstance


class Occurrence(BaseModel):
    hidden: bool
    fixed: bool
    transform: list[float]
    path: list[str]

    @property
    def world_to_part_tf(self) -> np.matrix:
        return np.matrix(np.array(self.transform).reshape(4, 4))

    @property
    def key(self) -> Key:
        return tuple(self.path)


class MateType(str, Enum):
    FASTENED = "FASTENED"
    REVOLUTE = "REVOLUTE"
    SLIDER = "SLIDER"
    PLANAR = "PLANAR"
    CYLINDRICAL = "CYLINDRICAL"
    PIN_SLOT = "PIN_SLOT"
    BALL = "BALL"
    PARALLEL = "PARALLEL"


class MatedCS(BaseModel):
    xAxis: list[float]
    yAxis: list[float]
    zAxis: list[float]
    origin: list[float]

    @property
    def part_to_mate_tf(self) -> np.matrix:
        part_to_mate_tf = np.eye(4)
        part_to_mate_tf[:3, :3] = np.stack(
            (
                np.array(self.xAxis),
                np.array(self.yAxis),
                np.array(self.zAxis),
            )
        ).T
        part_to_mate_tf[:3, 3] = self.origin
        return np.matrix(part_to_mate_tf)


class MatedEntity(BaseModel):
    matedOccurrence: list[str]
    matedCS: MatedCS

    @property
    def key(self) -> Key:
        return tuple(self.matedOccurrence)


class MateFeatureData(BaseModel):
    mateType: MateType
    matedEntities: list[MatedEntity]
    name: str


class RelationType(str, Enum):
    GEAR = "GEAR"
    LINEAR = "LINEAR"


class MateRelationMate(BaseModel):
    featureId: str
    occurrence: list[str]

    @property
    def key(self) -> Key:
        return tuple(self.occurrence + [self.featureId])


class MateRelationFeatureData(BaseModel):
    relationType: RelationType
    mates: list[MateRelationMate]
    reverseDirection: bool
    relationRatio: float
    name: str


class MateRelationFeature(BaseModel):
    id: str
    suppressed: bool
    featureType: Literal["mateRelation"]
    featureData: MateRelationFeatureData

    def keys(self, root_key: Key = ()) -> list[Key]:
        return [root_key + mated.key for mated in self.featureData.mates]


class MateFeature(BaseModel):
    id: str
    suppressed: bool
    featureType: Literal["mate"]
    featureData: MateFeatureData

    def keys(self, root_key: Key = ()) -> list[Key]:
        return [root_key + mated.key for mated in self.featureData.matedEntities]


class MateGroupFeatureOccurrence(BaseModel):
    occurrence: list[str]


class MateGroupFeatureData(BaseModel):
    occurrences: list[MateGroupFeatureOccurrence]
    name: str


class MateGroupFeature(BaseModel):
    id: str
    suppressed: bool
    featureType: Literal["mateGroup"]
    featureData: MateGroupFeatureData


class Pattern(BaseModel):
    pass


class RootAssembly(BaseModel):
    occurrences: list[Occurrence]
    instances: list[Instance]
    patterns: list[Pattern]
    features: list[MateGroupFeature | MateRelationFeature | MateFeature]
    fullConfiguration: str
    configuration: str
    documentMicroversion: str
    documentId: str
    elementId: str

    @property
    def key(self) -> ElementUid:
        return ElementUid(
            self.documentId,
            self.documentMicroversion,
            self.elementId,
            "",
            self.fullConfiguration,
        )


class SubAssembly(BaseModel):
    instances: list[Instance]
    patterns: list[Pattern]
    features: list[MateGroupFeature | MateRelationFeature | MateFeature]
    fullConfiguration: str
    configuration: str
    documentMicroversion: str
    documentId: str
    elementId: str

    @property
    def key(self) -> ElementUid:
        return ElementUid(
            self.documentId,
            self.documentMicroversion,
            self.elementId,
            "",
            self.fullConfiguration,
        )


class AssemblyProperty(BaseModel):
    name: str
    value: Any


class AssemblyMetadata(BaseModel):
    jsonType: str
    elementType: int
    mimeType: str
    elementId: str
    properties: list[AssemblyProperty]
    href: str

    @functools.cached_property
    def property_map(self) -> dict[str, Any]:
        return {prop.name: prop.value for prop in self.properties}


class PartBodyType(str, Enum):
    sheet = "sheet"
    solid = "solid"


class Part(BaseModel):
    isStandardContent: bool
    partId: str
    bodyType: PartBodyType
    fullConfiguration: str
    configuration: str
    documentMicroversion: str
    documentId: str
    elementId: str

    @property
    def key(self) -> ElementUid:
        return ElementUid(
            self.documentId,
            self.documentMicroversion,
            self.elementId,
            self.partId,
            self.fullConfiguration,
        )


class PartStudioFeature(BaseModel):
    pass


class Assembly(BaseModel):
    rootAssembly: RootAssembly
    subAssemblies: list[SubAssembly]
    parts: list[Part]
    partStudioFeatures: list[PartStudioFeature]



"""Defines a pass to convert the collision mesh into a convex hull."""

import argparse
import logging
import xml.etree.ElementTree as ET
from pathlib import Path

from onshape.formats.common import save_xml
from onshape.utils.mesh import COLLISION_SUFFIX

logger = logging.getLogger(__name__)


def make_convex_collision_mesh(
    mesh_path: Path,
    output_mesh_path: Path,
    max_triangles: int | None = None,
) -> tuple[int, int]:
    try:
        import open3d as o3d
    except ImportError:
        raise ImportError(
            "Open3D is required to run this script. Install it with `pip install "
            "'onshape[open3d]'` to install the required dependencies."
        )

    mesh = o3d.io.read_triangle_mesh(str(mesh_path))
    hull, _ = mesh.compute_convex_hull()
    hull = hull.remove_duplicated_vertices()
    hull = hull.remove_degenerate_triangles()
    hull = hull.remove_duplicated_triangles()

    # Remove triangles in order to greatly simplify the mesh.
    if max_triangles is not None and len(hull.triangles) > max_triangles:
        hull = hull.simplify_quadric_decimation(target_number_of_triangles=max_triangles)
        hull = hull.remove_duplicated_vertices()
        hull = hull.remove_degenerate_triangles()
        hull = hull.remove_duplicated_triangles()

    pre_num_vertices = len(mesh.vertices)
    post_num_vertices = len(hull.vertices)

    match output_mesh_path.suffix.lower():
        case ".ply":
            o3d.io.write_triangle_mesh(str(output_mesh_path), hull, write_ascii=True)
        case ".stl":
            hull.compute_vertex_normals()
            o3d.io.write_triangle_mesh(str(output_mesh_path), hull)
        case _:
            o3d.io.write_triangle_mesh(str(output_mesh_path), hull)

    if post_num_vertices < pre_num_vertices:
        logger.info(
            "Convex hull reduces meshes from %d to %d vertices (%.2f%% reduction) for %s",
            pre_num_vertices,
            post_num_vertices,
            (1 - post_num_vertices / pre_num_vertices) * 100,
            output_mesh_path,
        )

    return pre_num_vertices, post_num_vertices


def get_convex_collision_meshes(urdf_path: Path, max_triangles: int | None = None) -> None:
    """Converts the collision mesh into convex hulls.

    Args:
        urdf_path: The path to the urdf file.
        max_triangles: The maximum number of triangles to use for simplifying
            the collision meshes.

    Returns:
        The path to the merged urdf file.
    """
    urdf_tree = ET.parse(urdf_path)

    def get_mesh(visual_or_collision: ET.Element) -> tuple[ET.Element, Path] | None:
        if (geometry := visual_or_collision.find("geometry")) is None:
            return None
        if (mesh := geometry.find("mesh")) is None:
            return None
        return mesh, (urdf_path.parent / mesh.attrib["filename"]).resolve()

    # Iterates through each link in the URDF and simplifies the meshes.
    total_pre_num_vertices = 0
    total_post_num_vertices = 0
    has_change = False
    for link in urdf_tree.iter("link"):
        if (visual_link := link.find("visual")) is None or (collision_link := link.find("collision")) is None:
            continue
        if (visual_mesh := get_mesh(visual_link)) is None or (collision_mesh := get_mesh(collision_link)) is None:
            continue
        _, visual_mesh_path = visual_mesh
        collision_mesh_elem, collision_mesh_path = collision_mesh

        # If the visual mesh and the collision mesh are the same, we need to
        # make a new mesh for the collision mesh. Otherwise, we can just
        # overwrite the existing collision mesh.
        if visual_mesh_path == collision_mesh_path:
            new_suffix = f"{COLLISION_SUFFIX}{collision_mesh_path.suffix}"
            collision_output_mesh_path = collision_mesh_path.with_suffix(new_suffix)
            collision_mesh_elem.attrib["filename"] = collision_output_mesh_path.relative_to(urdf_path.parent).as_posix()
            has_change = True
        else:
            collision_output_mesh_path = collision_mesh_path

        pre_vertices, post_vertices = make_convex_collision_mesh(
            collision_mesh_path,
            collision_output_mesh_path,
            max_triangles=max_triangles,
        )
        total_pre_num_vertices += pre_vertices
        total_post_num_vertices += post_vertices

    if total_pre_num_vertices > 0:
        percent_reduction = (1 - total_post_num_vertices / total_pre_num_vertices) * 100
        logger.info(
            "Convex hull reduces meshes from %d to %d vertices (%.2f%% reduction)",
            total_pre_num_vertices,
            total_post_num_vertices,
            percent_reduction,
        )

    if has_change:
        save_xml(urdf_path, urdf_tree)


def main() -> None:
    parser = argparse.ArgumentParser(description="Merge fixed joints in a URDF.")
    parser.add_argument("urdf_path", type=Path, help="The path to the URDF file.")
    args = parser.parse_args()

    get_convex_collision_meshes(args.urdf_path)


if __name__ == "__main__":
    # python -m onshape.passes.simplify_meshes
    main()



"""Defines a pass to reduce the number of vertices in each mesh."""

import argparse
import logging
from pathlib import Path

from onshape.passes.utils import iter_meshes

logger = logging.getLogger(__name__)


def simplify_mesh(
    mesh_path: Path,
    voxel_size: float,
    max_triangles: int | None = None,
) -> tuple[int, int]:
    try:
        import open3d as o3d
    except ImportError:
        raise ImportError(
            "Open3D is required to run this script. Install it with `pip install "
            "'onshape[open3d]'` to install the required dependencies."
        )

    mesh = o3d.io.read_triangle_mesh(str(mesh_path))
    simple_mesh = mesh.simplify_vertex_clustering(
        voxel_size=voxel_size,
        contraction=o3d.geometry.SimplificationContraction.Average,
    )

    # Remove triangles in order to greatly simplify the mesh.
    if max_triangles is not None and len(simple_mesh.triangles) > max_triangles:
        simple_mesh = simple_mesh.simplify_quadric_decimation(target_number_of_triangles=max_triangles)
        simple_mesh = simple_mesh.remove_duplicated_vertices()
        simple_mesh = simple_mesh.remove_degenerate_triangles()
        simple_mesh = simple_mesh.remove_duplicated_triangles()

    pre_num_vertices = len(mesh.vertices)
    post_num_vertices = len(simple_mesh.vertices)

    match mesh_path.suffix.lower():
        case ".ply":
            o3d.io.write_triangle_mesh(str(mesh_path), simple_mesh, write_ascii=True)
        case ".stl":
            simple_mesh.compute_vertex_normals()
            o3d.io.write_triangle_mesh(str(mesh_path), simple_mesh)
        case _:
            o3d.io.write_triangle_mesh(str(mesh_path), simple_mesh)

    return pre_num_vertices, post_num_vertices


def get_simplified_urdf(
    urdf_path: Path,
    voxel_size: float = 0.002,
    max_triangles: int | None = None,
) -> None:
    """Merges meshes at each fixed joints to avoid collision issues.

    Args:
        urdf_path: The path to the urdf file.
        voxel_size: The voxel size to use for simplifying the meshes.
        max_triangles: The maximum number of triangles to use for simplifying
            the collision meshes.

    Returns:
        The path to the merged urdf file.
    """
    # Iterates through each link in the URDF and simplifies the meshes.
    for _, (_, visual_mesh_path), (_, collision_mesh_path) in iter_meshes(urdf_path):
        for mesh_path in list({visual_mesh_path, collision_mesh_path}):
            if mesh_path is not None:
                pre_num_vertices, post_num_vertices = simplify_mesh(
                    mesh_path,
                    voxel_size=voxel_size,
                    max_triangles=max_triangles,
                )

                if pre_num_vertices > post_num_vertices:
                    percent_reduction = (1 - post_num_vertices / pre_num_vertices) * 100
                    logger.info(
                        "Simplified meshes from %d to %d vertices (%.2f%% reduction) for %s",
                        pre_num_vertices,
                        post_num_vertices,
                        percent_reduction,
                        mesh_path,
                    )


def main() -> None:
    parser = argparse.ArgumentParser(description="Merge fixed joints in a URDF.")
    parser.add_argument("urdf_path", type=Path, help="The path to the URDF file.")
    parser.add_argument("--voxel-size", type=float, default=0.002, help="The voxel size for simplification.")
    args = parser.parse_args()

    get_simplified_urdf(args.urdf_path, voxel_size=args.voxel_size)


if __name__ == "__main__":
    # python -m onshape.passes.simplify_meshes
    main()



"""Removes any extra mesh files in the URDF directory."""

import argparse
import logging
from pathlib import Path

from onshape.passes.utils import iter_meshes

logger = logging.getLogger(__name__)


def remove_extra_meshes(urdf_path: Path) -> None:
    """Removes any extra mesh files in the URDF directory."""
    urdf_path = urdf_path.resolve()
    valid_paths = {urdf_path}
    mesh_dirs: set[Path] = set()
    for _, (_, visual_mesh_path), (_, collision_mesh_path) in iter_meshes(urdf_path):
        for path in list({visual_mesh_path, collision_mesh_path}):
            if path is not None:
                valid_paths.add(path)
                mesh_dirs.add(path.parent)

    all_meshes = {path for mesh_dir in mesh_dirs for path in mesh_dir.glob("**/*.stl")}
    for path in all_meshes:
        if path.is_file() and path not in valid_paths:
            logger.info("Removing %s", path)
            path.unlink()


def main() -> None:
    parser = argparse.ArgumentParser(description="Removes any extra mesh files in the URDF directory.")
    parser.add_argument("urdf_path", type=Path, help="The path to the URDF file.")
    args = parser.parse_args()

    remove_extra_meshes(args.urdf_path)


if __name__ == "__main__":
    # python -m onshape.passes.remove_extra_meshes
    main()



"""Defines functions for updating joint properties in a URDF."""

import logging
import xml.etree.ElementTree as ET
from pathlib import Path

from onshape.formats.common import save_xml

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)


def update_joints(
    urdf_path: Path,
    name_dict: dict[str, str] | None,
    override_fixed: list[str] | None,
    override_limits: dict[str, str] | None,
    override_torques: dict[str, int] | None,
) -> None:
    # Load the URDF file.
    logger.info("Updating joints.")
    tree = ET.parse(urdf_path)
    root = tree.getroot()

    # Sort keys so we match joints to longest first.
    if name_dict:
        sorted_keys = sorted(name_dict.keys(), key=len, reverse=True)

    # Collect all joints in urdf and get their meshes.
    for joint in root.iter("joint"):
        joint_name = joint.attrib["name"]

        # Remove "joint_" from beginning of joint name.
        if joint_name.startswith("joint_"):
            joint_name = joint_name[6:]

        # Check if name_dict is not None and if the joint_name is a substring
        # of any key in name_dict.
        if name_dict:
            for key in sorted_keys:
                if key in joint_name:
                    joint.attrib["name"] = name_dict[key]
                    break

            # If joint has a mimic that needs to be updated as well.
            mimic = joint.find("mimic")
            if mimic is not None:
                mimic_name = mimic.attrib["joint"]
                for key in sorted_keys:
                    if key in mimic_name:
                        mimic.attrib["joint"] = name_dict[key]
                        break

        # Check if override_dict is not None and if the any element of override
        # is substring of joint_name then fix.
        if override_fixed and any(ov in joint_name for ov in override_fixed):
            joint.attrib["type"] = "fixed"

        if override_torques:
            for key in override_torques.keys():
                if key in joint_name:
                    limit = joint.find("limit")
                    if limit is not None:
                        limit.attrib["effort"] = str(override_torques[key])
                    break

    for joint in root.iter("joint"):
        joint_name = joint.attrib["name"]
        # Check if override_limits is not None and if any key in
        # override_limits is a substring of joint_name.
        # Note: we do this with the new joint names
        if override_limits:
            for key in override_limits.keys():
                if key in joint_name:
                    limit = joint.find("limit")
                    if limit:
                        limit.attrib["lower"] = str(override_limits[key][0])
                        limit.attrib["upper"] = str(override_limits[key][1])
                    break

    # Save the updated URDF to the same file.
    save_xml(urdf_path, tree)
